require "spec_helper"

RSpec.describe TextEditor::Buffer do
  subject { described_class.new(lines) }

  let(:lines) { %w(one two) }

  describe "#delete_char" do
    it "deletes char at line and column" do
      actual = subject.delete_char(0, 0)
      expect(actual.to_s).to eq("ne\ntwo")

      actual = subject.delete_char(0, 1)
      expect(actual.to_s).to eq("oe\ntwo")

      actual = subject.delete_char(1, 0)
      expect(actual.to_s).to eq("one\nwo")

      actual = subject.delete_char(1, 1)
      expect(actual.to_s).to eq("one\nto")
    end

    it "ignores columns out of bounds" do
      actual = subject.delete_char(0, 10)
      expect(actual.to_s).to eq("one\ntwo")
    end

    it "raises IndexError when line is out of bounds" do
      expect { subject.delete_char(10, 0) }.to raise_error(IndexError)
    end
  end

  describe "#delete_line" do
    it "deletes line at index" do
      actual = subject.delete_line(0)
      expect(actual.to_s).to eq("two")

      actual = subject.delete_line(1)
      expect(actual.to_s).to eq("one")
    end

    it "ignores out of bound indexes" do
      actual = subject.delete_line(3)
      expect(actual.to_s).to eq("one\ntwo")
    end

    it "does not mutate the original buffer" do
      subject.delete_line(0)
      expect(subject.to_s).to eq("one\ntwo")
    end
  end

  describe "#each_line" do
    it "delegates to lines" do
      expect(lines).to receive(:each)
      subject.each_line { |line| line }
    end
  end

  describe "#insert_char" do
    it "inserts char at index" do
      actual = subject.insert_char(0, 0, "x")
      expect(actual.to_s).to eq("xone\ntwo")

      actual = subject.insert_char(0, 1, "x")
      expect(actual.to_s).to eq("oxne\ntwo")

      actual = subject.insert_char(1, 0, "x")
      expect(actual.to_s).to eq("one\nxtwo")

      actual = subject.insert_char(1, 1, "x")
      expect(actual.to_s).to eq("one\ntxwo")
    end

    it "raises IndexError when line is out of bounds" do
      expect { subject.insert_char(10, 0, "x") }.to raise_error(IndexError)
    end

    it "raises IndexError when column is out of bounds" do
      expect { subject.insert_char(0, 10, "x") }.to raise_error(IndexError)
    end
  end

  describe "#insert_line" do
    it "inserts line at index" do
      actual = subject.insert_line(0)
      expect(actual.to_s).to eq("\none\ntwo")

      actual = subject.insert_line(1)
      expect(actual.to_s).to eq("one\n\ntwo")

      actual = subject.insert_line(2)
      expect(actual.to_s).to eq("one\ntwo\n")
    end

    it "optionally accepts content" do
      actual = subject.insert_line(1, "test")
      expect(actual.to_s).to eq("one\ntest\ntwo")
    end

    it "inserts at tail when index is out of bounds" do
      actual = subject.insert_line(5)
      expect(actual.to_s).to eq("one\ntwo\n")
    end
  end

  describe "#line" do
    it "returns line by index" do
      actual = subject.line(0)
      expect(actual).to eq("one")

      actual = subject.line(1)
      expect(actual).to eq("two")
    end

    it "returns empty string otherwise" do
      actual = subject.line(2)
      expect(actual).to eq("")
    end

    it "returns a copy of the line" do
      first_id = subject.line(0).object_id
      second_id = subject.line(0).object_id
      expect(first_id).not_to eq(second_id)
    end
  end

  context "#replace_char" do
    it "replace char at index" do
      actual = subject.replace_char(0, 0, "x")
      expect(actual.to_s).to eq("xne\ntwo")

      actual = subject.replace_char(0, 1, "x")
      expect(actual.to_s).to eq("oxe\ntwo")

      actual = subject.replace_char(1, 0, "x")
      expect(actual.to_s).to eq("one\nxwo")

      actual = subject.replace_char(1, 1, "x")
      expect(actual.to_s).to eq("one\ntxo")
    end

    it "raises IndexError when line is out of bounds" do
      expect { subject.replace_char(10, 0, "x") }.to raise_error(IndexError)
    end

    it "raises IndexError when column is out of bounds" do
      expect { subject.replace_char(0, 10, "x") }.to raise_error(IndexError)
    end
  end

  context "#replace_line" do
    it "replaces line at index" do
      actual = subject.replace_line(0, "test")
      expect(actual.to_s).to eq("test\ntwo")

      actual = subject.replace_line(1, "test")
      expect(actual.to_s).to eq("one\ntest")
    end

    it "replaces line with empty string by default" do
      actual = subject.replace_line(0)
      expect(actual.to_s).to eq("\ntwo")
    end

    it "raises IndexError when line is out of bounds" do
      expect { subject.replace_line(10, "test") }.to raise_error(IndexError)
    end
  end

  describe "#to_s" do
    it "joins lines with a newline" do
      expect(subject.to_s).to eq("one\ntwo")
    end

    it "returns an empty string if lines are empty" do
      lines.clear
      expect(subject.to_s).to eq("")
    end
  end
end
